@@ -3,14 +3,13 @@ Core consumer and event-loop code.
 """
 
 import logging
+import time
 
 from django.conf import settings
 from django.core.management.base import BaseCommand
-from django.dispatch import receiver
+from edx_django_utils.monitoring import record_exception
 from edx_toggles.toggles import SettingToggle
 from openedx_events.event_bus.avro.deserializer import AvroSignalDeserializer
-from openedx_events.learning.data import UserData
-from openedx_events.learning.signals import SESSION_LOGIN_COMPLETED
 from openedx_events.tooling import OpenEdxPublicSignal
 
 from .config import get_full_topic, get_schema_registry_client, load_common_settings
@@ -20,7 +19,7 @@ logger = logging.getLogger(__name__)
 # See https://github.com/openedx/event-bus-kafka/blob/main/docs/decisions/0005-optional-import-of-confluent-kafka.rst
 try:
     import confluent_kafka
-    from confluent_kafka import DeserializingConsumer, KafkaError
+    from confluent_kafka import DeserializingConsumer
     from confluent_kafka.schema_registry.avro import AvroDeserializer
 except ImportError:  # pragma: no cover
     confluent_kafka = None
@@ -36,11 +35,29 @@ KAFKA_CONSUMERS_ENABLED = SettingToggle('EVENT_BUS_KAFKA_CONSUMERS_ENABLED', def
 
 CONSUMER_POLL_TIMEOUT = getattr(settings, 'EVENT_BUS_KAFKA_CONSUMER_POLL_TIMEOUT', 1.0)
 
+# .. setting_name: EVENT_BUS_KAFKA_CONSUMER_POLL_FAILURE_SLEEP
+# .. setting_default: 1.0
+# .. setting_description: When the consumer fails to retrieve an event from the broker,
+#   it will sleep for this many seconds before trying again. This is to prevent fast error-loops
+#   if the broker is down or the consumer is misconfigured. It *may* also sleep for errors that
+#   involve receiving an unreadable event, but this could change in the future to be more
+#   specific to "no event received from broker".
+POLL_FAILURE_SLEEP = getattr(settings, 'EVENT_BUS_KAFKA_CONSUMER_POLL_FAILURE_SLEEP', 1.0)
+
 # CloudEvent standard name for the event type header, see
 # https://github.com/cloudevents/spec/blob/v1.0.1/kafka-protocol-binding.md#325-example
 EVENT_TYPE_HEADER = "ce_type"
 
 
+class UnusableMessageError(Exception):
+    """
+    Indicates that a message was successfully received but could not be processed.
+
+    This could be invalid headers, an unknown signal, or other issue specific to
+    the contents of the message.
+    """
+
+
 class KafkaEventConsumer:
     """
     Construct consumer for the given topic, group, and signal. The consumer can then
@@ -57,6 +74,7 @@ class KafkaEventConsumer:
         self.group_id = group_id
         self.signal = signal
         self.consumer = self._create_consumer()
+        self._shut_down_loop = False
 
     # return type (Optional[DeserializingConsumer]) removed from signature to avoid error on import
     def _create_consumer(self):
@@ -70,10 +88,6 @@ class KafkaEventConsumer:
 
         schema_registry_client = get_schema_registry_client()
 
-        # TODO (EventBus):
-        # 1. Reevaluate if all consumers should listen for the earliest unprocessed offset (auto.offset.reset)
-        # 2. Ensure the signal used in the signal_deserializer is the same one sent over in the message header
-
         signal_deserializer = AvroSignalDeserializer(self.signal)
 
         def inner_from_dict(event_data_dict, ctx=None):  # pylint: disable=unused-argument
@@ -88,11 +102,16 @@ class KafkaEventConsumer:
             'value.deserializer': AvroDeserializer(schema_str=signal_deserializer.schema_string(),
                                                    schema_registry_client=schema_registry_client,
                                                    from_dict=inner_from_dict),
-            'auto.offset.reset': 'earliest'
         })
 
         return DeserializingConsumer(consumer_config)
 
+    def _shut_down(self):
+        """
+        Test utility for shutting down the consumer loop.
+        """
+        self._shut_down_loop = True
+
     def consume_indefinitely(self):
         """
         Consume events from a topic in an infinite loop.
@@ -100,39 +119,42 @@ class KafkaEventConsumer:
         # This is already checked at the Command level, but it's possible this loop
         # could get called some other way, so check it here too.
         if not KAFKA_CONSUMERS_ENABLED.is_enabled():
-            logger.error("Kafka consumers not enabled")
+            logger.error("Kafka consumers not enabled, exiting.")
             return
 
         try:
             full_topic = get_full_topic(self.topic)
+            run_context = {
+                'full_topic': full_topic,
+                'consumer_group': self.group_id,
+                'expected_signal': self.signal,
+            }
             self.consumer.subscribe([full_topic])
+            logger.info(f"Running consumer for {run_context!r}")
 
-            # TODO (EventBus):
-            # 1. Is there an elegant way to exit the loop?
-            # 2. Determine if there are other errors that shouldn't kill the entire loop
             while True:
-                msg = self.consumer.poll(timeout=CONSUMER_POLL_TIMEOUT)
-                if msg is not None:
-                    self.process_single_message(msg)
+                # Allow unit tests to break out of loop
+                if self._shut_down_loop:
+                    break
+
+                msg = None
+                try:
+                    msg = self.consumer.poll(timeout=CONSUMER_POLL_TIMEOUT)
+                    if msg is not None:
+                        self.emit_signals_from_message(msg)
+                except BaseException as e:
+                    self.record_event_consuming_error(run_context, e, msg)
+                    # Prevent fast error-looping when no event received from broker. Because
+                    # DeserializingConsumer raises rather than returning a Message when it has an
+                    # error() value, this may be triggered even when a Message *was* returned,
+                    # slowing down the queue. This is probably close enough, though.
+                    if msg is None:
+                        time.sleep(POLL_FAILURE_SLEEP)
         finally:
             # Close down consumer to commit final offsets.
             self.consumer.close()
             logger.info("Committing final offsets")
 
-    def process_single_message(self, msg):
-        """
-        Emit signal with message data
-        """
-        if msg.error():
-            # TODO (EventBus): iterate on error handling with retry and dead-letter queue topics
-            if msg.error().code() == KafkaError._PARTITION_EOF:  # pylint: disable=protected-access
-                # End of partition event
-                logger.info(f"{msg.topic()} [{msg.partition()}] reached end at offset {msg.offset}")
-            else:
-                logger.exception(msg.error())
-        else:
-            self.emit_signals_from_message(msg)
-
     def emit_signals_from_message(self, msg):
         """
         Determine the correct signal and send the event from the message.
@@ -140,56 +162,85 @@ class KafkaEventConsumer:
         Arguments:
             msg (Message): Consumed message.
         """
+        # DeserializingConsumer.poll() always returns either a valid message
+        # or None, and raises an exception in all other cases. This means
+        # we don't need to check msg.error() ourselves. But... check it here
+        # anyway for robustness against code changes.
+        if msg.error() is not None:
+            raise UnusableMessageError(
+                f"Polled message had error object (shouldn't happen): {msg.error()!r}"
+            )
+
         headers = msg.headers() or []  # treat None as []
 
-        # TODO (EventBus): iterate on error handling for missing or multiple event_type headers
-        #  (headers() is actually a list of (key, value) tuples rather than a dictionary)
         event_types = [value for key, value in headers if key == EVENT_TYPE_HEADER]
         if len(event_types) == 0:
-            logger.error(f"Missing {EVENT_TYPE_HEADER} header on message, cannot determine signal")
-            return
+            raise UnusableMessageError(
+                f"Missing {EVENT_TYPE_HEADER} header on message, cannot determine signal"
+            )
         if len(event_types) > 1:
-            logger.error(f"Multiple {EVENT_TYPE_HEADER} headers found on message, cannot determine signal")
-            return
+            raise UnusableMessageError(
+                f"Multiple {EVENT_TYPE_HEADER} headers found on message, cannot determine signal"
+            )
 
         event_type = event_types[0]
 
-        # TODO (EventBus): Figure out who is doing the encoding and get the
-        #  right one instead of just guessing utf-8
+        # CloudEvents specifies using UTF-8 for header values, so let's be explicit.
         event_type_str = event_type.decode("utf-8")
+
+        if event_type_str != self.signal.event_type:
+            raise UnusableMessageError(
+                f"Signal types do not match. Expected {self.signal.event_type}. "
+                f"Received message of type {event_type_str}."
+            )
+
+        self.signal.send_event(**msg.value())
+
+    def record_event_consuming_error(self, run_context, error, maybe_event):
+        """
+        Record an error caught while consuming an event, both to the logs and to telemetry.
+
+        Arguments:
+            run_context: Dictionary of contextual information: full_topic, consumer_group,
+              and expected_signal.
+            error: An exception instance
+            maybe_event: None if event could not be fetched or decoded, or a Message if one
+              was successfully deserialized but could not be processed for some reason
+        """
+        context_msg = ", ".join(f"{k}={v!r}" for k, v in run_context.items())
+        if maybe_event is None:
+            event_msg = "no event available"
+        else:
+            event_details = {
+                'partition': maybe_event.partition(),
+                'offset': maybe_event.offset(),
+                'headers': maybe_event.headers(),
+                'key': maybe_event.key(),
+                'value': maybe_event.value(),
+            }
+            event_msg = f"event details: {event_details!r}"
+
         try:
-
-            # If we get a message with the wrong signal encoding, we do not want to send it along.
-            # TODO (EventBus): Handle this particular sad path more gracefully.
-            if event_type_str != self.signal.event_type:
-                logger.error(
-                    f"Signal types do not match. Expected {self.signal.event_type}."
-                    f"Received message of type {event_type_str}."
-                )
-                return
-
-            signal = OpenEdxPublicSignal.get_signal_by_type(event_type_str)
-            signal.send_event(**msg.value())
-        except KeyError:
-            logger.exception(f"Signal not found: {event_type_str}")
+            # This is gross, but our record_exception wrapper doesn't take args at the moment,
+            # and will only read the exception from stack context.
+            raise Exception(error)
+        except BaseException:
+            record_exception()
+            logger.exception(
+                f"Error consuming event from Kafka: {error!r} in context {context_msg} -- {event_msg}"
+            )
 
 
 class ConsumeEventsCommand(BaseCommand):
     """
-    Listen for events from the event bus and log them. Only run on servers where
-    ``EVENT_BUS_KAFKA_CONSUMERS_ENABLED`` is true.
+    Management command for Kafka consumer workers in the event bus.
     """
     help = """
-    This starts a Kafka event consumer that listens to the specified topic and logs all messages it receives. Topic
-    is required.
+    Consume messages of specified signal type from a Kafka topic and send their data to that signal.
 
     example:
-        python3 manage.py cms consume_events -t user-event-debug -g user-event-consumers
+        python3 manage.py cms consume_events -t user-login -g user-activity-service \
             -s org.openedx.learning.auth.session.login.completed.v1
-
-    # TODO (EventBus): Add pointer to relevant future docs around topics and consumer groups, and potentially
-    update example topic and group names to follow any future naming conventions.
-
     """
 
     def add_arguments(self, parser):
@@ -198,7 +249,7 @@ class ConsumeEventsCommand(BaseCommand):
             '-t', '--topic',
             nargs=1,
             required=True,
-            help='Topic to consume'
+            help='Topic to consume (without environment prefix)'
         )
 
         parser.add_argument(
@@ -216,7 +267,7 @@ class ConsumeEventsCommand(BaseCommand):
 
     def handle(self, *args, **options):
         if not KAFKA_CONSUMERS_ENABLED.is_enabled():
-            logger.error("Kafka consumers not enabled")
+            logger.error("Kafka consumers not enabled, exiting.")
             return
         try:
             signal = OpenEdxPublicSignal.get_signal_by_type(options['signal'][0])
@@ -228,24 +279,3 @@ class ConsumeEventsCommand(BaseCommand):
             event_consumer.consume_indefinitely()
         except Exception:  # pylint: disable=broad-except
             logger.exception("Error consuming Kafka events")
-
-
-@receiver(SESSION_LOGIN_COMPLETED)
-def log_event_from_event_bus(**kwargs):  # pragma: no cover
-    """
-    Log event received and transmitted from event bus consumer.
-
-    This is test code that should be removed.
-
-    Arguments:
-        kwargs: event data sent to the signal
-    """
-    try:
-        user_data = kwargs.get('user', None)
-        if not user_data or not isinstance(user_data, UserData):
-            logger.error("Received null or incorrect data from SESSION_LOGIN_COMPLETED")
-            return
-        logger.info(f"Received SESSION_LOGIN_COMPLETED signal with user_data"
-                    f" with UserData {user_data}")
-    except Exception:  # pylint: disable=broad-except
-        logger.exception("Error while testing receiving signals from Kafka events")
